#include <cilk-internal.h>

#if CILK_FILEIO

#include "fileio.h"
#include <errno.h>

/*
 * We use file descriptor integers to talk about files.  These
 * are small integers returned by the open() routine.  The system
 * reserves 0, 1, and 2 for stdin, stdout, and stderr.  We also
 * reserve these values for the same purpose.
 */
#define LAST_SPECIAL_FILE 2
#define FIRST_USER_FILE (LAST_SPECIAL_FILE+1)

typedef struct
{
  int os_fd;
  off_t offset;
  int clean;
  off_t offset_after_sync;
  int index_after_sync;
} FileEntryT;

typedef struct FileTableT
{
  FileEntryT entry[CILK_MAX_FILES];
  LockT lock;
  /* int max_valid; - should use this eventually -KHR */
} FileTableT;

/*
 * Note: need to obtain lock on FileTableT only to play with
 * *_after_sync entries.  All others are locally owned by
 * the processor on which the table is installed.
 */

/* Cilk files */
DECLARE_PRIVATE_VARIABLE(static FileTableT *local_table) = NULL;

typedef struct {
     char *path;
     int flags;
     /* other stuff, as we get more I/O crazy */
} FileT;

DECLARE_SHARED_VARIABLE(static FileT file_table[CILK_MAX_FILES]);
DECLARE_SHARED_VARIABLE(static AlignedLockT file_table_lock);

/*
 * Note: file_table_lock is used only when allocating a cilk file
 * descriptor ID.
 */

static int make_os_fd(FileEntryT *entry, int cilk_fd)
{
  int os_fd;
  int res;

  CILK_ASSERT(entry->os_fd < 0);
  CILK_ASSERT(cilk_fd >= 0);
  
  if (cilk_fd <= LAST_SPECIAL_FILE)
    {
      os_fd = cilk_fd;
    }
  else
    {
      CILK_CHECK(file_table[cilk_fd].path != NULL,
		 ("bad cilk file ID %d\n", cilk_fd));	/* return
							 * EBADF? */
      os_fd = open(file_table[cilk_fd].path, file_table[cilk_fd].flags);
      CILK_CHECK(os_fd >= 0,
		 ("couldn't translate file %s\n", file_table[cilk_fd].path));
    }

  res = lseek(os_fd, entry->offset, SEEK_SET);
  if ((res == -1 && errno == ESPIPE) || isatty(os_fd))
    {
      /* this file is not seekable. */
      /* Cilk_die("cilk can't operate on unseekable files!\n"); */
    }
  else
    {
      CILK_ASSERT(res == entry->offset);
    }

  /* remember translation for later */
  entry->os_fd = os_fd;
  
  TRACE(TRACE_FILES,
	("opened file %d (local %d) at %d\n", cilk_fd, os_fd, entry->offset));
  return os_fd;
}

/***********************************************************\
 * Unix file operations
\***********************************************************/

#include <stdarg.h>

/*
 * Open call.
 */
int Cilk_open(const char *path, int flags, ... /* int mode */ )
{
     int cilk_fd, os_fd;
     va_list l;
     int mode;
     FileEntryT *entry;
     char *save_path = Cilk_malloc(strlen(path) + 1);
     strcpy(save_path, path);

     /* get mode */
     va_start(l, flags);
     mode = va_arg(l, typeof(mode));
     va_end(l);

     /* get file table lock */
     Cilk_sema_wait(&file_table_lock.lock);

     /* find a free Cilk file ID */
     for (cilk_fd = FIRST_USER_FILE; cilk_fd < CILK_MAX_FILES; cilk_fd++)
	  if (file_table[cilk_fd].path == NULL)
	       break;
     if (cilk_fd == CILK_MAX_FILES)
       {
	 errno = EMFILE;
	 Cilk_sema_signal(&file_table_lock.lock);
	 Cilk_free(save_path);
	 return -1;
       }

     file_table[cilk_fd].path = save_path;

     /* release file table lock */
     Cilk_sema_signal(&file_table_lock.lock);

     /* set up file locally */
     CILK_ASSERT(local_table != NULL);
     entry = &local_table->entry[cilk_fd];
     CILK_ASSERT(entry->os_fd == -1);
     os_fd = open(path, flags, mode);
     if (os_fd < 0) {
	  file_table[cilk_fd].path = NULL;
	  Cilk_membar_StoreStore();
	  Cilk_free(save_path);
	  return os_fd;		/* open returned an error - give it to
				 * user */
     }
     entry->os_fd = os_fd;
     entry->offset = 0;
     entry->clean = 0;
     entry->offset_after_sync = -1;
     entry->index_after_sync = 0;

     /* 
      * Note: we need to use some flags only once when opening the
      * file, like the O_TRUNC flag.  We perform the open here,
      * with all flags given, and then remove the flags that should
      * only be used once.
      *
      * Currently, some flags are just bad, like the O_APPEND flag.
      * For now, assert that any flags we don't know about aren't
      * included.
      *
      * O_RDONLY, O_WRONLY, O_RDWR : OK
      * O_TRUNC, O_CREAT, O_EXCL   : should be used only once.
      * O_APPEND                   : bad
      * many others                : potentially bad, don't allow them for now.
      */

     /* remove any flags that are no longer needed */
     flags &= ~(O_TRUNC | O_CREAT | O_EXCL);

     /* for now, check to make sure there aren't other weird flags */
     CILK_ASSERT((flags & ~(O_RDONLY | O_WRONLY | O_RDWR)) == 0);

     /* save rest of flags for local opens */
     file_table[cilk_fd].flags = flags;

     TRACE(TRACE_FILES,
	   ("first open of file %d (local %d)\n", cilk_fd, os_fd));

     return cilk_fd;
}

/*
 * Close call.
 */
int Cilk_close(int cilk_fd)
{
  int os_fd;
  FileEntryT *entry;

  CILK_ASSERT(local_table != NULL);
  entry = &local_table->entry[cilk_fd];
  
  /* close local os file, if any */
  os_fd = entry->os_fd;
  if (os_fd >= 0)
    {
      CILK_CHECK(close(os_fd) == 0,
		 ("couldn't close local file in Cilk_close(%d)\n", cilk_fd));
      entry->os_fd = -1;
      TRACE(TRACE_FILES,
	    ("last close of file %d (local %d)\n", cilk_fd, os_fd));
    }
  /* release cilk file id */
  if (cilk_fd >= FIRST_USER_FILE)
    {
      Cilk_free(file_table[cilk_fd].path);
      file_table[cilk_fd].path = NULL;
      Cilk_membar_StoreStore();
    }
  return 0;
}

/*
 * Read <len> bytes from location <offset> in file <cilk_fd> into buffer <buf>.
 */
int Cilk_read(int cilk_fd, void *buf, size_t len)
{
     FileEntryT *entry;
     int os_fd;
     int res;

     CILK_ASSERT(local_table != NULL);
     entry = &local_table->entry[cilk_fd];
     if ((os_fd = entry->os_fd) < 0)
       os_fd = make_os_fd(entry, cilk_fd);

     /* see Cilk_write for a comment */
     WHEN_CILK_DEBUG(res = lseek(os_fd, 0, SEEK_CUR));
     CILK_ASSERT(res == entry->offset ||
		 (res == -1 && errno == ESPIPE) ||
		 isatty(os_fd));

     Cilk_touch_mem(buf, len);
     res = read(os_fd, buf, len);
     if (res == -1)
          return res;

     entry->offset += res;
     entry->clean = 0;
     return res;
}

/*
 * Write <len> bytes to location <offset> in file <cilk_fd> from buffer <buf>.
 */
int Cilk_write(int cilk_fd, void *buf, size_t len)
{
     FileEntryT *entry;
     int os_fd;
     int res;

     CILK_ASSERT(local_table != NULL);
     entry = &local_table->entry[cilk_fd];
     if ((os_fd = entry->os_fd) < 0)
       os_fd = make_os_fd(entry, cilk_fd);

     /*
      * Why don't we get an ESPIPE error on a tty file descriptor?
      * lseek works, but the file pointer is never used.  Weird...
      */
     WHEN_CILK_DEBUG(res = lseek(os_fd, 0, SEEK_CUR));
     CILK_ASSERT(res == entry->offset ||
		 (res == -1 && errno == ESPIPE) ||
		 isatty(os_fd));

     Cilk_touch_mem(buf, len);
     res = write(os_fd, buf, len);
     if (res == -1)
          return res;

     entry->offset += res;
     entry->clean = 0;
     return res;
}

off_t Cilk_lseek(int cilk_fd, off_t offset, int whence)
{
     FileEntryT *entry;
     int os_fd;
     off_t res;

     CILK_ASSERT(local_table != NULL);
     entry = &local_table->entry[cilk_fd];
     if ((os_fd = entry->os_fd) < 0)
       os_fd = make_os_fd(entry, cilk_fd);

     res = lseek(os_fd, offset, whence);
     if (res == -1)
          return res;

     entry->offset = res;
     entry->clean = 0;
     return res;
}


/***********************************************************\
 * Callback routines used by the rts
\***********************************************************/

FileTableT *Cilk_fileio_empty_table(void)
{
  int i;
  FileTableT *new = (FileTableT*)Cilk_internal_malloc(sizeof(FileTableT));
  Cilk_sema_init(&new->lock);
  for(i=FIRST_USER_FILE; i<CILK_MAX_FILES; i++)
    {
      new->entry[i].os_fd = -1;
      new->entry[i].clean = 1;
    }
  
  /* pretend stdin, stdout, stderr have already been opened */
  for(i=0; i<=LAST_SPECIAL_FILE; i++)
    {
      new->entry[i].os_fd = -1;
      new->entry[i].offset = 0;
      new->entry[i].clean = 0;
      new->entry[i].offset_after_sync = -1;
      new->entry[i].index_after_sync = 0;
    }
  return new;
}

void Cilk_fileio_free_table(FileTableT *table)
{
  CILK_ASSERT(table != NULL);
  Cilk_internal_free(table, sizeof(FileTableT));
}

void Cilk_fileio_start_hook(volatile SubcomputationT *subcomp)
{
  CILK_ASSERT(local_table == NULL);
  local_table = subcomp->file_table;
#if CILK_DEBUG
  {
    /* make sure none of the cilk files have matching OS files yet. */
    int i;
    for(i=0; i<CILK_MAX_FILES; i++)
      {
	CILK_ASSERT(local_table->entry[i].os_fd == -1);
      }
  }
#endif
}

void Cilk_fileio_stop_hook(volatile SubcomputationT *subcomp)
{
  int cilk_fd;
  int os_fd;
  FileTableT *table;

  table = local_table;
  CILK_ASSERT(table != NULL);
  CILK_ASSERT(table == subcomp->file_table);
  WHEN_CILK_DEBUG(local_table = NULL);
  
  /* close all local os file descriptors */
  for (cilk_fd = 0; cilk_fd < CILK_MAX_FILES; cilk_fd++)
    {
      os_fd = table->entry[cilk_fd].os_fd;
      if (os_fd >= 0)
	{
	  if (cilk_fd >= FIRST_USER_FILE)   /* don't close stdin,stdout,stderr */
	    {
	      close(os_fd);
	      TRACE(TRACE_FILES,
		    ("closed file %d (local %d)\n", cilk_fd, os_fd));
	    }
	  table->entry[cilk_fd].os_fd = -1;
	}
    }
}

/*
 * This should be called only when caller has a lock on <parent>.
 * <child> should be newly-allocated.
 */
void Cilk_fileio_steal_hook(volatile SubcomputationT *child,
			    volatile SubcomputationT *parent)
{
  int i;
  FileTableT *old = parent->file_table;
  FileTableT *new = (FileTableT*)Cilk_internal_malloc(sizeof(FileTableT));

  CILK_ASSERT(old != NULL);
  Cilk_sema_init(&new->lock);
  Cilk_sema_wait(&old->lock);
  for(i=0; i<CILK_MAX_FILES; i++)
    {
      FileEntryT *newentry = &new->entry[i];
      FileEntryT *oldentry = &old->entry[i];
      
      newentry->os_fd = -1;
      newentry->offset = oldentry->offset;
      newentry->clean = 1;
      newentry->offset_after_sync = oldentry->offset_after_sync;
      newentry->index_after_sync = oldentry->index_after_sync;

      oldentry->offset_after_sync = -1;
      oldentry->index_after_sync = 0;
    }
  parent->file_table = new;
  child->file_table = old;
  Cilk_sema_signal(&old->lock);
}

/* slow sync only */
void Cilk_fileio_sync_hook(volatile SubcomputationT *subcomp)
{
  int i;
  FileTableT *table = local_table;
  FileEntryT *entry;
  CILK_ASSERT(table != NULL);
  CILK_ASSERT(table == subcomp->file_table);
  for(i=0; i<CILK_MAX_FILES; i++)
    {
      entry = &table->entry[i];
      if (entry->clean)
	if (entry->index_after_sync > 0)
	  {
	    CILK_ASSERT(entry->offset_after_sync >= 0);
	    TRACE(TRACE_FILES,
		  ("at sync: file %d old %d new %d from child %d\n",
		   i,
		   entry->offset,
		   entry->offset_after_sync,
		   entry->index_after_sync));
	    entry->offset = entry->offset_after_sync;
	    entry->clean = 0;
	    entry->offset_after_sync = -1;
	    entry->index_after_sync = 0;
	    
	    if (entry->os_fd != -1)
	      lseek(entry->os_fd, entry->offset, SEEK_SET);
	  }
    }
}

/*
 * Called from returnit, used to update parent file structures when
 * child returns to it.
 */
void Cilk_fileio_return_hook(volatile SubcomputationT *child,
			     volatile SubcomputationT *parent)
{
  int cilk_fd;
  FileTableT *child_table, *parent_table;
  FileEntryT *child_entry, *parent_entry;

  child_table = child->file_table;
  CILK_ASSERT(child_table != NULL);
  parent_table = parent->file_table;
  CILK_ASSERT(parent_table != NULL);

  /* copy dirty file pointers from child to parent */
  Cilk_sema_wait(&parent_table->lock);
  for(cilk_fd = 0; cilk_fd < CILK_MAX_FILES; cilk_fd++)
    {
      child_entry = &child_table->entry[cilk_fd];
      parent_entry = &parent_table->entry[cilk_fd];
      if (!child_entry->clean)
	if (child->parent_index > parent_entry->index_after_sync)
	{
	    TRACE(TRACE_FILES,
		  ("at return: file %d installing %d of child %d over %d of child %d\n",
		   cilk_fd,
		   child_entry->offset,
		   child->parent_index,
		   parent_entry->offset_after_sync,
		   parent_entry->index_after_sync));
	  parent_entry->index_after_sync = child->parent_index;
	  parent_entry->offset_after_sync = child_entry->offset;
	}
    }
  Cilk_sema_signal(&parent_table->lock);
  
  Cilk_internal_free(child_table, sizeof(FileTableT));
  WHEN_CILK_DEBUG(child->file_table = NULL);
}

/***********************************************************\
 * Initialization
\***********************************************************/

static void fileio_init_global_hook(void)
{
  int i;
  Cilk_sema_init(&file_table_lock.lock);
  for(i=0; i<CILK_MAX_FILES; i++)
    file_table[i].path = NULL;
}

static void init_fileio(void) __attribute__((constructor));
static void init_fileio(void)
{
  Cilk_add_hook(&Cilk_initial_global_hooks, fileio_init_global_hook);
}

#endif				/* CILK_FILEIO */
