/*
 * $Id: bstore.c,v 1.4 1996/09/05 19:21:41 randall Exp $
 */

/*
 * $Log: bstore.c,v $
 * Revision 1.4  1996/09/05 19:21:41  randall
 * Small changes.
 *
 * Revision 1.3  1996/08/27 19:28:17  randall
 * One pass of cleaning up.
 *
 * Revision 1.2  1996/08/27 15:43:21  athena
 * Some more cleanup -- changed some low-level interfaces.
 *
 * Revision 1.1  1996/08/26 22:16:37  athena
 * Major restructuring of the rts. Added shared and private variables.
 * Added hooks. Added cilklib.
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/***********************************************************
 * Backing store
 ***********************************************************/
#include <cilk-internal.h>

/* Cilk_bstore_mem: size of backing-store file */
DECLARE_SHARED_VARIABLE(size_t Cilk_bstore_mem) = CILK_DEFAULT_BSTORESIZE;

DECLARE_SHARED_VARIABLE(static volatile int bs_free_list);
DECLARE_SHARED_VARIABLE(static volatile int *bs_flist_array);
WHEN_CILK_DEBUG(DECLARE_SHARED_VARIABLE(static volatile int bs_used_pages);)
DECLARE_SHARED_VARIABLE(static AlignedLockT bs_lock);

void Cilk_bstore_init(void)
{
     int npages;		/* number of pages */
     int i;

     TRACE(TRACE_SETUP, ("Start setup bstore\n"));

     Cilk_bstore_mem = ROUND_2_PAGE(Cilk_bstore_mem);
     CILK_CHECK(PAGE_ALIGNED_P(Cilk_bstore_mem),
		("Cilk_bstore_mem not multiple of Cilk_pagesize!\n"));

     Cilk_setup_bstore(Cilk_bstore_mem);

     npages = Cilk_bstore_mem / Cilk_pagesize;

     /* allocate a chunk of shared memory to hold the free list */
     bs_flist_array = Cilk_malloc(npages * sizeof(int));

     /* init the free list to contain all pages */
     for (i = 0; i < npages; ++i) {
	  bs_flist_array[i] = i + 1;
     }

     bs_flist_array[npages - 1] = -1;

     /* init the lock */
     Cilk_sema_init(&bs_lock.lock);

     /* initialize the free list pointer */
     bs_free_list = 0;		/* first free page */

     /* count # of used pages */
     WHEN_CILK_DEBUG(bs_used_pages = 0);

     TRACE(TRACE_SETUP, ("End setup bstore\n"));
}

int Cilk_alloc_bstore_page(void)
{
     int page;

     Cilk_enter_state(STATE_ALLOC_BSTORE_PAGE);

     Cilk_enter_state(STATE_WAIT_BS_LOCK);
     Cilk_sema_wait(&bs_lock.lock);
     Cilk_exit_state(STATE_WAIT_BS_LOCK);

     page = bs_free_list;

     CILK_CHECK(page != -1,
		("out of bstore pages!\n"));

     bs_free_list = bs_flist_array[page];

     WHEN_CILK_DEBUG(bs_used_pages++);

     /* 
      * This check ensures that we don't use the same page twice.
      */
     CILK_ASSERT(bs_flist_array[page] != -2);
     WHEN_CILK_DEBUG(bs_flist_array[page] = -2);

     Cilk_sema_signal(&bs_lock.lock);

     CILK_ASSERT(page >= 0 && page < Cilk_bstore_mem / Cilk_pagesize);
     Cilk_exit_state(STATE_ALLOC_BSTORE_PAGE);
     return page;
}

void Cilk_free_bstore_page(int page)
{
     Cilk_enter_state(STATE_FREE_BSTORE_PAGE);

     CILK_ASSERT(page >= 0 && page < Cilk_bstore_mem / Cilk_pagesize);

     Cilk_enter_state(STATE_WAIT_BS_LOCK);
     Cilk_sema_wait(&bs_lock.lock);
     Cilk_exit_state(STATE_WAIT_BS_LOCK);

     bs_flist_array[page] = bs_free_list;
     bs_free_list = page;

     WHEN_CILK_DEBUG(bs_used_pages--);

     Cilk_sema_signal(&bs_lock.lock);
     Cilk_exit_state(STATE_FREE_BSTORE_PAGE);
}

#if CILK_DEBUG
void Cilk_check_bstore(void)
{
     if (bs_used_pages != 0)
	  CILK_DPRINTF("%d pages of backing store still allocated.\n",
		       bs_used_pages);
}
#endif
