/*
 * $Id: cilk-cilk2c.h,v 1.62 1996/09/10 16:20:29 athena Exp $
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <cilk-conf.h>

/*
 * We define Cilk_ss here so we can use the const declaration for
 * Cilk_ss in the user code, while Cilk_ss in the run-time system
 * is not const.  Quite a hack, but it works.
 */
extern volatile struct StealStuffT *const Cilk_ss;

/* add current frame to deque */
#define _PUSH_FRAME(f) { \
    WHEN_CILK_DEBUG(f->header.magic = CILK_FRAME_STEALABLE_MAGIC); \
    Cilk_membar_StoreStore(); \
    Cilk_ss->tail = (StackFrameT*)((char*)f+f->header.size); \
}

#define _ALIGN_SIZE(s) ((s+CILK_MAX_ALIGN-1) & (~(CILK_MAX_ALIGN-1)))

/* predicates */
#define _INIT_FRAME(f,s,signat) { \
     f = (void*)(Cilk_ss->tail); \
     f->header.size = _ALIGN_SIZE(s); \
     WHEN_CILK_DEBUG(f->header.magic = CILK_FRAME_NONSTEALABLE_MAGIC); \
     f->header.sig = (signat); \
}

#define _CHILD_VALID(f) !Cilk_result_valid
#define _SET_RESULT(x) *((typeof(x)*)Cilk_result) = x
#define _SET_VALID Cilk_result_valid = 1
#define _SYNC Cilk_sync((StackFrameT*)_frame)

#define _HEAPIFY(f) CILK_ASSERT(!Cilk_result_valid)

#define _RESUME() Cilk_resume = 1
#define _ABORT_SLOW() Cilk_abort_from_slow()
#define _ABORT_STANDALONE() Cilk_abort = 1

/* Dekker-like protocol, see steal() in cilk.c for details */
#define _XPOP_FRAME_RESULT(f, r, result) \
{ \
     Cilk_ss->tail = (StackFrameT *)f; \
     Cilk_membar_StoreLoad(); \
     if (Cilk_ss->head > (StackFrameT *)f) { \
	  if (Cilk_stolen_check((StackFrameT *)f)) { \
	       if (_CHILD_VALID(f)) { _SET_RESULT(result); _SET_VALID; } \
	       return r ; \
	  } \
     } \
     CILK_ASSERT(f->header.magic == CILK_FRAME_STEALABLE_MAGIC); \
     WHEN_CILK_DEBUG(f->header.magic = CILK_FRAME_NONSTEALABLE_MAGIC); \
}

#define _XPOP_FRAME_NORESULT(f, r) \
{ \
     Cilk_ss->tail = (StackFrameT *)f; \
     Cilk_membar_StoreLoad(); \
     if (Cilk_ss->head > (StackFrameT *)f) { \
	  if (Cilk_stolen_check((StackFrameT *)f)) { \
	       if (_CHILD_VALID(f)) { _SET_VALID; } \
	       return r ; \
	  } \
     } \
     CILK_ASSERT(f->header.magic == CILK_FRAME_STEALABLE_MAGIC); \
     WHEN_CILK_DEBUG(f->header.magic = CILK_FRAME_NONSTEALABLE_MAGIC); \
}

#define offsetof(type, mem) ((size_t) ((char *)&((type *) 0)->mem - (char *)((type *) 0)))

#if CILK_CRITICAL_PATH
/*
 * Critical-path measurements. Each procedure has two variables,
 * CP and MYCP.
 *
 * Let a "nice node" be the beginning of the thread, or any sync
 * point. CP and MYCP agree at nice nodes, and are the actual critical
 * path up to that node.
 *
 * On non-nice nodes, CP and MYCP have different meanings.
 * MYCP is the critical path up to the last nice node, plus the
 * execution time since then. CP is the maximum critical path
 * of all children spawned and finished since last nice node.
 *
 * Therefore, when a child is spawned, it starts with the current
 * value of MYCP; moreover, when sync occurs, CP = MYCP = max(CP, MYCP).
 *
 * The variable Cilk_cp_hack is used to pass CP from parent to
 * child, and vice versa.
 *
 * CP and MYCP for the fast version are local variables, saved on the
 * frame before a spawn. The slow version always operates on the frame.
 *
 * We also take advantage of the fact the CP measurements are slow
 * to collect a fine-grained timing of the total work at no appreciable
 * cost. Every thread as a variable WORK that collects the relevant
 * quantities.  The variable Cilk_work_hack passes this value
 * from child to parent (the vice versa is not needed).
 */

#if CILK_STATS
extern void Cilk_stat_broken_thread(void);
extern void Cilk_stat_new_thread(void);
extern void Cilk_stat_null_thread(void);
#endif

extern const char *Cilk_assertion_failed_msg; /* for CILK_ASSERT */
extern void Cilk_die(const char *fmt,...);

#if CILK_CRITICAL_PATH == CILK_CP_PROCESS
#define CILK_GET_TIME() gethrvtime()
#define CILK_UPDATE_TIME(a) a
static inline long long CILK_GET_TIME_INTERVAL(long long start)
{
     extern long long CILK_GET_TIME();
     long long c = CILK_GET_TIME();

     /* this is really paranoid */
     CILK_ASSERT(c >= start);

     WHEN_CILK_STATS(if (c == start) Cilk_stat_null_thread());
     return c;
}
#endif

#if CILK_CRITICAL_PATH == CILK_CP_WALL
#define CILK_GET_TIME() gethrtime()
#define CILK_UPDATE_TIME(a) a
static inline long long CILK_GET_TIME_INTERVAL(long long start)
{
     extern long long CILK_GET_TIME();
     long long c = CILK_GET_TIME();

     /* this is really paranoid */
     CILK_ASSERT(c >= start);

     WHEN_CILK_STATS(if (c == start) Cilk_stat_null_thread());
     return c;
}
#endif

#if CILK_CRITICAL_PATH == CILK_CP_BROKEN_THREAD
#define CILK_GET_TIME() gethrtime()
#define CILK_UPDATE_TIME(a) gethrtime()
static inline long long CILK_GET_TIME_INTERVAL(long long start)
{
     long long c;
     extern long long gethrtime();

     c = gethrtime();
     /* this is really paranoid */
     CILK_ASSERT(c >= start);

     WHEN_CILK_STATS(if (c == start) Cilk_stat_null_thread());
     if (c > start + CILK_BROKEN_THREAD_THRESHOLD) {
	  WHEN_CILK_STATS(Cilk_stat_broken_thread());
	  c = start;
     }
     return c;

}
#endif

/* at begin of thread, set my critical path to the one of the parent */
#define _START_THREAD_FAST() \
      TimeT Cilk_cp = 0, Cilk_mycp = Cilk_cp_hack, \
      Cilk_begin_time = CILK_GET_TIME(), Cilk_work = 0;\
      WHEN_CILK_STATS(Cilk_stat_new_thread());

/* 
 * but the slow version by definition has a valid CP and MYCP, so
 * no initialization is needed.
 */
#define _START_THREAD_SLOW() \
      TimeT Cilk_begin_time = CILK_GET_TIME();

#define _AT_THREAD_BOUNDARY_SLOW() \
      WHEN_CILK_STATS(Cilk_stat_new_thread());

/*
 * before spawn, adjust mycp and pass it to the child (implicitly) 
 */
#define _BEFORE_SPAWN_FAST() {\
      TimeT Cilk_curr_time = CILK_GET_TIME_INTERVAL(Cilk_begin_time); \
      Cilk_cp_hack = (Cilk_mycp += (Cilk_curr_time -  Cilk_begin_time));\
      Cilk_work += (Cilk_curr_time -  Cilk_begin_time);\
      _frame->header.mycp = Cilk_mycp;\
      _frame->header.cp = Cilk_cp;\
      _frame->header.work = Cilk_work;\
      Cilk_begin_time = CILK_UPDATE_TIME(Cilk_curr_time); \
}

#define _BEFORE_SPAWN_SLOW() {\
      TimeT Cilk_curr_time = CILK_GET_TIME_INTERVAL(Cilk_begin_time); \
      Cilk_cp_hack = (_frame->header.mycp +=  \
              (Cilk_curr_time -  Cilk_begin_time));\
      _frame->header.work +=  \
              (Cilk_curr_time -  Cilk_begin_time);\
      Cilk_begin_time = CILK_UPDATE_TIME(Cilk_curr_time); \
}

/* after a spawn, reset Cilk_begin_time, maximize CP and accumulate WORK */
#define _AFTER_SPAWN_FAST()  {\
   Cilk_begin_time = CILK_GET_TIME();\
   WHEN_CILK_STATS(Cilk_stat_new_thread());\
   if (Cilk_cp_hack > Cilk_cp) Cilk_cp = Cilk_cp_hack;\
   Cilk_work += Cilk_work_hack; \
}

#define _AFTER_SPAWN_SLOW()  {\
   Cilk_begin_time = CILK_GET_TIME();\
   if (Cilk_cp_hack > _frame->header.cp) _frame->header.cp = Cilk_cp_hack;\
   _frame->header.work += Cilk_work_hack;\
}

/* at a sync, set MYCP = CP = max(CP, MYCP) and accumulate work */
#define _AT_SYNC_FAST() {\
      TimeT Cilk_curr_time = CILK_GET_TIME_INTERVAL(Cilk_begin_time); \
      Cilk_mycp += (Cilk_curr_time - Cilk_begin_time);\
      Cilk_work += (Cilk_curr_time - Cilk_begin_time);\
      if (Cilk_mycp < Cilk_cp) \
	   Cilk_mycp = Cilk_cp;\
      else \
	   Cilk_cp = Cilk_mycp;\
      WHEN_CILK_STATS(Cilk_stat_new_thread());\
      Cilk_begin_time = CILK_UPDATE_TIME(Cilk_curr_time); \
}

/*
 * the slow case is trickier. Before the check for sync,
 * we update CP/MYCP and work.
 *
 * *After* the check we maximize CP with all CP's of children
 * executed by other processors. This value is in turn stored
 * in the subcomp structure, and can be read without race conditions because
 * there are no oustanding children after a sync (well, almost.
 * This is broken in case of resume).
 * Also we accumulate the work with the total work of child
 * subcomputations.
 */
#define _BEFORE_SYNC_SLOW() {\
      TimeT Cilk_curr_time = CILK_GET_TIME_INTERVAL(Cilk_begin_time); \
      _frame->header.mycp += (Cilk_curr_time - Cilk_begin_time);\
      _frame->header.work += (Cilk_curr_time - Cilk_begin_time);\
      if (_frame->header.cp > _frame->header.mycp) \
	   _frame->header.mycp = _frame->header.cp;\
      else \
	   _frame->header.cp = _frame->header.mycp;\
      Cilk_begin_time = CILK_UPDATE_TIME(Cilk_curr_time); \
}

#define _AFTER_SYNC_SLOW() {\
       if (Cilk_ss->subcomp->cp > _frame->header.mycp) \
	    _frame->header.cp = _frame->header.mycp = Cilk_ss->subcomp->cp;\
       _frame->header.work += Cilk_ss->subcomp->work;\
       Cilk_ss->subcomp->work = 0;\
}

/* return is like sync, but also sets CP for the parent */
#define _BEFORE_RETURN_FAST() {\
      TimeT Cilk_curr_time = CILK_GET_TIME_INTERVAL(Cilk_begin_time); \
      Cilk_mycp += (Cilk_curr_time - Cilk_begin_time);\
      Cilk_work += (Cilk_curr_time - Cilk_begin_time);\
      if (Cilk_cp > Cilk_mycp) \
	   Cilk_cp_hack = Cilk_cp;\
      else \
	   Cilk_cp_hack = Cilk_mycp;\
      Cilk_work_hack = Cilk_work;\
}

/* no need to check subcomp here, since there is a sync before return */
#define _BEFORE_RETURN_SLOW() {\
      TimeT Cilk_curr_time = CILK_GET_TIME_INTERVAL(Cilk_begin_time); \
      _frame->header.mycp += (Cilk_curr_time -  Cilk_begin_time);\
      _frame->header.work += (Cilk_curr_time -  Cilk_begin_time);\
      if (_frame->header.cp > _frame->header.mycp) \
	   Cilk_cp_hack = _frame->header.cp;\
      else \
	   Cilk_cp_hack = _frame->header.mycp;\
      Cilk_work_hack = _frame->header.work;\
}

#define _BEFORE_RETURN_INLET()

#else				/* !CILK_CRITICAL_PATH */

#define _BEFORE_SPAWN_FAST()

#define _BEFORE_SPAWN_SLOW() _BEFORE_SPAWN_FAST()
#define _AT_THREAD_BOUNDARY_SLOW() 
#define _AFTER_SPAWN_FAST()
#define _AFTER_SPAWN_SLOW() _AFTER_SPAWN_FAST()

#define _START_THREAD_FAST()
#define _START_THREAD_SLOW() _START_THREAD_FAST()

#define _BEFORE_RETURN_FAST()
#define _BEFORE_RETURN_SLOW() _BEFORE_RETURN_FAST()

#define _BEFORE_RETURN_INLET()

#define _BEFORE_SYNC_SLOW()
#define _AFTER_SYNC_SLOW()
#define _AT_SYNC_FAST()
#endif
