/* $Id: cilk-internal.h,v 1.11 1996/09/16 03:29:47 fengmd Exp $ */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <cilk.h>
#include <cilk-lib.h>

/* common definitions used internally by the rts */

/* backing store */
extern void Cilk_bstore_init(void);
extern int Cilk_alloc_bstore_page(void);
extern void Cilk_free_bstore_page(int page);
WHEN_CILK_DEBUG(extern void Cilk_check_bstore(void);)

/* internal malloc */
extern void Cilk_internal_malloc_init(void);
extern volatile void *Cilk_internal_malloc(unsigned size);
extern void Cilk_internal_free(volatile void *area, int size);
WHEN_CILK_DEBUG(extern void Cilk_check_internal_malloc(void);)

/* command-line parser */
extern int Cilk_parse_command_line(int *argc, char *argv[]);

/* file io */
#if CILK_FILEIO
struct FileTableT *Cilk_fileio_empty_table(void);
void Cilk_fileio_free_table(struct FileTableT *table);

void Cilk_fileio_start_hook(volatile SubcomputationT *subcomp);
void Cilk_fileio_stop_hook(volatile SubcomputationT *subcomp);
void Cilk_fileio_steal_hook(volatile SubcomputationT *child,
			    volatile SubcomputationT *parent);
void Cilk_fileio_sync_hook(volatile SubcomputationT *subcomp);
void Cilk_fileio_return_hook(volatile SubcomputationT *child,
			     volatile SubcomputationT *parent);
#endif

/* other variables */
DECLARE_SHARED_VARIABLE(extern size_t Cilk_bstore_mem);
DECLARE_SHARED_VARIABLE(extern size_t Cilk_pagesize);
DECLARE_SHARED_VARIABLE(extern int CilkStatLevel);
DECLARE_SHARED_VARIABLE(extern void *Cilk_frame_stack);
DECLARE_SHARED_VARIABLE(extern size_t Cilk_proc_mem);

/* external malloc */
DECLARE_SHARED_VARIABLE(extern size_t Cilk_arenasize);
extern void Cilk_init_arena();

/* callbacks from architecture-specific files */
extern int Cilk_signal_handler(void *fault_addr);

#define ROUND_2_PAGE(addr) (((addr) + Cilk_pagesize - 1) / Cilk_pagesize \
			    * Cilk_pagesize)

#define PAGE_ALIGNED_P(addr) (((int) (addr)) % Cilk_pagesize == 0)

/* from page # to stack address */
#define PAGE_2_STACK_ADDR(page) ((char *)Cilk_frame_stack + (page) * Cilk_pagesize)

/* from stack address to page # */
#define STACK_ADDR_2_PAGE(addr) (((char *)(addr) - (char *)Cilk_frame_stack) / Cilk_pagesize)

/* from size in pages to size in bytes */
#define PAGE_2_BYTES(pages) ((pages) * Cilk_pagesize)


/* statistics */
#if CILK_STATS

typedef struct {
     CACHE_LINE_ALIGNED(int steal_attempts);
     int steals;
     volatile int page_faults;
     long nthreads;
     long broken_threads;
     long null_threads;
} StatsT;

extern void Cilk_stat_steal_attempt(void);
extern void Cilk_stat_steal(void);
extern void Cilk_stat_page_fault(void);
extern void Cilk_stats_gathering_init(void);
extern void Cilk_print_rts_statistics(void);
extern void Cilk_stat_broken_thread(void);
extern void Cilk_stat_new_thread(void);
extern void Cilk_stat_null_thread(void);

#else

#define Cilk_stat_steal_attempt()
#define Cilk_stat_steal()
#define Cilk_stat_page_fault()
#define Cilk_stats_gathering_init()
#define Cilk_print_rts_statistics()
#define Cilk_stat_broken_thread()
#define Cilk_stat_new_thread()
#define Cilk_stat_null_thread()

#endif

/* timing */
#if CILK_TIMING
extern void Cilk_enter_state(int state);
extern void Cilk_exit_state(int state);
extern void Cilk_print_time_statistics(void);
extern void Cilk_print_detailed_time_statistics(void);
extern double Cilk_compute_work(void);
extern void Cilk_time_gathering_init(void);

enum {
     STATE_TOTAL,
     STATE_COPYING_PAGE_TABLE,
     STATE_FREEING_PAGE_TABLE,
     STATE_STEALING,
     STATE_GET_PAGE_ENTRY,
     STATE_ALLOC_BSTORE_PAGE,
     STATE_FREE_BSTORE_PAGE,
     STATE_WAIT_BS_LOCK,
     STATE_ALLOC_SHMEM,
     STATE_FREE_SHMEM,
     STATE_MAPPING_PAGE,
     STATE_WORKING,
     TIME_NSTATES		/* this must be last */
};

#else

#define Cilk_enter_state()
#define Cilk_exit_state()
#define Cilk_print_time_statistics()
Cilk_print_detailed_time_statistics()
#define Cilk_compute_work() (0.0)
#define Cilk_time_gathering_init()

#endif				/* timing */

extern void Cilk_page_table_init(void);
extern volatile struct PageTableT *Cilk_copy_page_table(volatile struct PageTableT *table,
							int page);
extern int Cilk_get_page_table_entry(volatile struct PageTableT *table,
				     int page);
extern volatile struct PageTableT *Cilk_dup_page_table(volatile struct PageTableT *table);
extern void Cilk_free_page_table(volatile struct PageTableT *table);
extern volatile struct PageTableT *Cilk_create_empty_page_table(void);
extern int Cilk_page_tables_cmp(volatile struct PageTableT *t1,
				volatile struct PageTableT *t2);
extern void Cilk_invalidate_stack_pages(int first, int last);

/* Hooks */
typedef void (*HookT)(void);

typedef struct hook {
     HookT fn;
     struct hook *next;
} HookList;

extern void Cilk_add_hook(HookList **listp, HookT fn);
extern void Cilk_run_hooks(HookList *list);
extern void Cilk_init_hooks(void);

#define NULL_HOOK ((HookList *) 0)

/*
 * various types of hooks 
 * This list is incomplete and should be updated as needed
 */
extern HookList *Cilk_initial_global_hooks;
extern HookList *Cilk_initial_local_hooks;
extern HookList *Cilk_return_hooks;

#if CILK_DEBUG
/* debugging support */
extern void Cilk_dump_frame(StackFrameT *f);
extern void Cilk_dump_subcomputation(SubcomputationT *s);
#endif

#if CILK_GDB
extern void send_message(unsigned int message);
extern void send_cilk_rts_info();
extern void send_end_message();
extern void set_up_tty_and_pause(int cilk_process_num);
#endif
