/*
 * $Id: cmdline.c,v 1.3 1996/09/17 01:24:51 athena Exp $
 */

/*
 * $Log: cmdline.c,v $
 * Revision 1.3  1996/09/17 01:24:51  athena
 * Added -infofile option
 *
 * Revision 1.2  1996/08/27 19:28:28  randall
 * One pass of cleaning up.
 *
 * Revision 1.1  1996/08/26 22:16:54  athena
 * Major restructuring of the rts. Added shared and private variables.
 * Added hooks. Added cilklib.
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/***********************************************************\
 * Command line parsing
\***********************************************************/

#include <cilk-internal.h>
#include <stdio.h>

DECLARE_SHARED_VARIABLE(extern char *Cilk_infofile_name);

enum {
     NONE, NPROC, STATS, NO_STATS, HELP, DEBUGOPT, ARENA,
     STACK, BSTORE_SIZE, INFOFILE, END_OPTIONS
};

static struct options {
     char *string;
     int option;
} optarray[] = {

     {
	  "-", END_OPTIONS	/* -- forces end to cilk options */
     },
     {
	  "nproc", NPROC
     },
     {
	  "stats", STATS
     },
     {
	  "no-stats", NO_STATS
     },
     {
	  "help", HELP
     },
     {
	  "debug", DEBUGOPT
     },
     {
	  "heap", ARENA
     },
     {
	  "stack", STACK
     },
     {
	  "bsize", BSTORE_SIZE
     },
     {
	  "info", INFOFILE
     },
     {
	  "infofile", INFOFILE
     },
     {
	  (char *) 0, NONE
     }
};

static void print_help(void)
{
     fprintf(stderr, "Cilk options:\n");
     fprintf(stderr, "  -nproc <number of processors>\n");
     fprintf(stderr, "  -stats <statistics level>\n");
     fprintf(stderr, "  -no-stats\n");
     fprintf(stderr, "  -debug <trace mode>\n");
     fprintf(stderr, "  -bfile <backing store file>\n");
     fprintf(stderr, "  -heap <size in MB>\n");
     fprintf(stderr, "  -stack <size in MB>\n");
     fprintf(stderr, "  -bsize <size in MB>\n");
     fprintf(stderr, "  -infofile <filename>\n");
     fprintf(stderr, "  -help\n");
     fprintf(stderr, "Use `--' to force Cilk option parsing to stop.\n");
}

/* look for a given string in the option table */
static int parse_option(char *s)
{
     struct options *p;

     for (p = optarray; p->string; ++p)
	  if (strcmp(s, p->string) == 0)
	       break;

     return p->option;
}

int Cilk_parse_command_line(int *argc, char *argv[])
{
     int i, j;
     for (i = 1; i < *argc; ++i) {
	  if (*argv[i] != '-')
	       goto arch_specific;
	  switch (parse_option(argv[i] + 1)) {
	      case NPROC:
		   ++i;
		   CILK_CHECK(i < *argc,
			      ("-nproc requires a number\n"));
		   Cilk_active_size = atoi(argv[i]);
		   if (Cilk_active_size == 0)
			Cilk_active_size = Cilk_partition_size();
		   CILK_CHECK(Cilk_active_size > 0,
			      ("non-positive number of processors\n"));
		   break;
	      case STATS:
		   ++i;
		   CILK_CHECK(i < *argc,
			      ("-stats requires a number\n"));
		   CilkStatLevel = atoi(argv[i]);
		   break;
	      case NO_STATS:
		   CilkStatLevel = 0;
		   break;
	      case DEBUGOPT:
		   ++i;
		   CILK_CHECK(i < *argc,
			      ("-debug requires a number\n"));
		   Cilk_trace_mode = strtol(argv[i], (char **) 0, 0);
		   break;
	      case HELP:
		   print_help();
		   return 1;
		   break;
	      case ARENA:
		   ++i;
		   CILK_CHECK(i < *argc,
			      ("%s requires a number\n", argv[i - 1]));
		   Cilk_arenasize = atof(argv[i]) * MEGABYTE;
		   break;
	      case STACK:
		   ++i;
		   CILK_CHECK(i < *argc,
			      ("%s requires a number\n", argv[i - 1]));
		   Cilk_proc_mem = atof(argv[i]) * MEGABYTE;
		   break;
	      case BSTORE_SIZE:
		   ++i;
		   CILK_CHECK(i < *argc,
			      ("%s requires a number\n", argv[i - 1]));
		   Cilk_bstore_mem = atof(argv[i]) * MEGABYTE;
		   break;
	      case INFOFILE:
		   ++i;
		   CILK_CHECK(i < *argc,
			      ("%s requires a file name\n", argv[i - 1]));
		   Cilk_infofile_name = argv[i];
		   break;
	      case END_OPTIONS:
		   ++i;
		   goto stop;
	      default:
		   goto arch_specific;
	  }
     }

     /* let architecture-specific code get a crack at the command line */
   arch_specific:
     i = Cilk_arch_specific_options(i, *argc, argv);

     /* remove Cilk command-line arguments from argv */
   stop:
     for (j = i; j <= *argc; j++)
	  argv[j - (i - 1)] = argv[j];
     *argc -= i - 1;

     return 0;
}
