/*
 * $Id: internal-malloc.c,v 1.4 1996/09/06 19:26:13 randall Exp $
 */

/*
 * $Log: internal-malloc.c,v $
 * Revision 1.4  1996/09/06 19:26:13  randall
 * Some small updates.  Added file I/O with file pointers.  All file I/O
 * stuff is #ifdef'd out for now.
 *
 * Revision 1.3  1996/09/05 19:21:54  randall
 * Small changes.
 *
 * Revision 1.2  1996/08/27 19:28:30  randall
 * One pass of cleaning up.
 *
 * Revision 1.1  1996/08/26 22:16:57  athena
 * Major restructuring of the rts. Added shared and private variables.
 * Added hooks. Added cilklib.
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <cilk-internal.h>

/***********************************************************\
 * Shared memory heap for use by system
\***********************************************************/

/* 
 * currently, only page tables and subcomputations are allocated
 * using this mechanism.
 */

/* 
 * Allocate just under a power of two so that Cilk_malloc doesn't
 * waste space.
 */
#define SHMEM_SIZE (255*CACHE_LINE)

DECLARE_SHARED_VARIABLE(static void *shmem_end);
	       /* end of shared memory heap */

DECLARE_SHARED_VARIABLE(static void *shmem_free);
    /* free pointer for shared memory  kept cache-line aligned */

DECLARE_SHARED_VARIABLE(static AlignedLockT shmem_lock);
    /* lock on shared-memory allocator */

WHEN_CILK_DEBUG(DECLARE_SHARED_VARIABLE(static int shmem_count);)
				/* amount of allocated memory */

#define MAX_SMALL_ALLOC (2048 / CACHE_LINE)

DECLARE_SHARED_VARIABLE(
       static volatile void *volatile free_shmem_array[MAX_SMALL_ALLOC]);

/*
 * Set up internal shared-memory heap.  
 * Call this once before processes are forked.  
 */
void Cilk_internal_malloc_init(void)
{
     int i;
     void *shmem_area;

     shmem_area = Cilk_malloc(SHMEM_SIZE);

     TRACE(TRACE_SETUP,
	   ("got shared-memory area %x\n", (unsigned int) shmem_area));

     shmem_end = shmem_area + SHMEM_SIZE;
     shmem_free = NEXT_CACHE_LINE(shmem_area);
     Cilk_sema_init(&shmem_lock.lock);

     for (i = 0; i < MAX_SMALL_ALLOC; i++)
	  free_shmem_array[i] = NULL;

     /* count amount of allocated shared memory */
     WHEN_CILK_DEBUG(shmem_count = 0);
}

/*
 * System routine for allocating from the system shared memory heap.
 * Used for subcomputations, and page tables.  Returned blocks are
 * CACHE_LINE aligned.
 */
volatile void *Cilk_internal_malloc(unsigned size)
{
     volatile void *area;

     Cilk_enter_state(STATE_ALLOC_SHMEM);
     size = (size + CACHE_LINE - 1) >> LOG_CACHE_LINE;
     CILK_ASSERT(size < MAX_SMALL_ALLOC);
     CILK_ASSERT(size > 0);

     Cilk_sema_wait(&shmem_lock.lock);
     WHEN_CILK_DEBUG(shmem_count += size);

     /* reuse small items */
     if ((area = free_shmem_array[size]) != NULL) {
	  free_shmem_array[size] = *((volatile void **) area);
	  Cilk_sema_signal(&shmem_lock.lock);
	  Cilk_exit_state(STATE_ALLOC_SHMEM);
	  return area;
     }
     area = shmem_free;
     shmem_free += (size << LOG_CACHE_LINE);

     if (shmem_free > shmem_end) {
	  /* need to Cilk_malloc some more memory! */
	  void *shmem_area = Cilk_malloc(SHMEM_SIZE);

	  TRACE(TRACE_SETUP,
	     ("got more shared-memory %x\n", (unsigned int) shmem_area));

	  shmem_end = shmem_area + SHMEM_SIZE;
	  shmem_free = NEXT_CACHE_LINE(shmem_area);
	  area = shmem_free;
	  shmem_free += (size << LOG_CACHE_LINE);
     }
     CILK_ASSERT(shmem_free <= shmem_end);

     Cilk_sema_signal(&shmem_lock.lock);
     Cilk_exit_state(STATE_ALLOC_SHMEM);
     return area;
}

void Cilk_internal_free(volatile void *area, int size)
{
     Cilk_enter_state(STATE_FREE_SHMEM);
     size = (size + CACHE_LINE - 1) >> LOG_CACHE_LINE;
     CILK_ASSERT(size < MAX_SMALL_ALLOC);
     CILK_ASSERT(size > 0);

     Cilk_sema_wait(&shmem_lock.lock);
     WHEN_CILK_DEBUG(shmem_count -= size);
     *((volatile void **) area) = free_shmem_array[size];
     free_shmem_array[size] = area;
     Cilk_sema_signal(&shmem_lock.lock);

     Cilk_exit_state(STATE_FREE_SHMEM);
}

#if CILK_DEBUG
void Cilk_check_internal_malloc(void)
{
     if (shmem_count != 0)
	  CILK_DPRINTF("%d cache lines of shared memory not freed.\n",
		       shmem_count);
}
#endif
