/*
 * $Id: page-table.c,v 1.2 1996/08/27 19:28:32 randall Exp $
 */

/*
 * $Log: page-table.c,v $
 * Revision 1.2  1996/08/27 19:28:32  randall
 * One pass of cleaning up.
 *
 * Revision 1.1  1996/08/26 22:16:59  athena
 * Major restructuring of the rts. Added shared and private variables.
 * Added hooks. Added cilklib.
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <cilk-internal.h>

/***********************************************************\
 * Page table
\***********************************************************/

/*
 * PAGE_BITS = # of bits that specify a page.  For 32-bit addressing
 * with a page size of 8KB, this gives 19 bits.
 */
#define PAGE_BITS 19

/*
 * LEAF_BITS = # of bits used to index leaf nodes.
 */
#define LEAF_BITS 4

/*
 * NODE_BITS = # of bits used to index internal nodes.
 */
#define NODE_BITS 3

/*
 * LEVELS = # of internal levels of table.
 */
#define NODE_LEVELS 5

/*
 *  We must have PAGE_BITS = LEAF_BITS + NODE_BITS * NODE_LEVELS.
 */

#define MAX_PAGE (1<<PAGE_BITS)
#define LEAF_SIZE (1<<LEAF_BITS)
#define NODE_SIZE (1<<NODE_BITS)

/*
 *  Structure of internal page table nodes.
 */
typedef struct PageTableT {
     LockT lock;
     int refcount;		/* number of incoming pointers */

     volatile void *child[NODE_SIZE];
} PageTableT;

/*
 *  Structure of leaf nodes.
 */
typedef struct {
     LockT lock;
     int refcount;		/* number of incoming pointers */

     int min_valid;		/* minimum valid index */
     volatile void *lower_half;	/* where to go for invalid pages */
     int pages[LEAF_SIZE];
} PageBaseT;

void Cilk_page_table_init(void)
{
     CILK_CHECK(PAGE_BITS == LEAF_BITS + NODE_BITS * NODE_LEVELS,
		("wrong definition of PAGE_BITS\n"));
}

/* 
 * Get base node holding mapping for page <page>-z.  <table> represents
 * pages z through z+<max>-1.
 */
static volatile PageBaseT *get_base(volatile PageTableT *node,
				    int max, int page)
{
     volatile PageTableT *child;
     int step, idx, off;
     int i;

     CILK_ASSERT(node != NULL);
     CILK_ASSERT(page >= 0);
     CILK_ASSERT(page < max);

     if (max == LEAF_SIZE)
	  return (PageBaseT *) node;

     step = max / NODE_SIZE;	/* # of pages for each child */
     idx = page / step;		/* child # of relevant child */
     off = page % step;		/* index in relevant child  */

     child = node->child[idx];

     if (child == NULL) {
	  Cilk_sema_wait(&node->lock);
	  child = node->child[idx];

	  if (child == NULL) {
	       if (step == LEAF_SIZE) {
		    /* allocate new base node and set it up */
		    volatile PageBaseT *base = Cilk_internal_malloc(sizeof(PageBaseT));
		    Cilk_sema_init(&base->lock);
		    base->refcount = 1;
		    base->min_valid = 0;
		    base->lower_half = NULL;
		    for (i = 0; i < LEAF_SIZE; i++)
			 base->pages[i] = -1;

		    child = (PageTableT *) base;
	       } else {
		    /* allocate new table node and set it up */
		    child = Cilk_internal_malloc(sizeof(PageTableT));
		    Cilk_sema_init(&child->lock);
		    child->refcount = 1;
		    for (i = 0; i < NODE_SIZE; i++)
			 child->child[i] = NULL;
	       }
	       node->child[idx] = child;
	  }
	  Cilk_sema_signal(&node->lock);
     }
     CILK_ASSERT(child != NULL);

     return (get_base(child, step, off));
}

int Cilk_get_page_table_entry(volatile PageTableT *table, int page)
{
     volatile PageBaseT *base;
     int idx;
     int bs_page;

     Cilk_enter_state(STATE_GET_PAGE_ENTRY);
     CILK_ASSERT(table);
     CILK_ASSERT(table->refcount > 0);
     CILK_ASSERT(page < MAX_PAGE);

     base = get_base(table, MAX_PAGE, page);
     idx = page % LEAF_SIZE;

     while (idx < base->min_valid)
	  base = base->lower_half;

     bs_page = base->pages[idx];
     if (bs_page == -1) {
	  Cilk_sema_wait(&base->lock);
	  bs_page = base->pages[idx];
	  if (bs_page == -1)
	       base->pages[idx] = bs_page = Cilk_alloc_bstore_page();
	  Cilk_sema_signal(&base->lock);
     }
     Cilk_exit_state(STATE_GET_PAGE_ENTRY);
     return bs_page;
}

/* inc_refcount and dec_refcount work on both PageTableTs and PageBaseTs */
static void inc_refcount(volatile PageTableT *table)
{
     Cilk_sema_wait(&table->lock);
     table->refcount++;
     Cilk_sema_signal(&table->lock);
}

static void dec_refcount(volatile PageTableT *table, int max)
{
     int i;

     Cilk_sema_wait(&table->lock);
     table->refcount--;
     CILK_ASSERT(table->refcount >= 0);
     if (table->refcount == 0) {
	  Cilk_sema_signal(&table->lock);
	  if (max == LEAF_SIZE) {
	       /* free PageBaseT structure */
	       volatile PageBaseT *base = (PageBaseT *) table;
	       for (i = base->min_valid; i < LEAF_SIZE; i++) {
		    if (base->pages[i] != -1)
			 Cilk_free_bstore_page(base->pages[i]);
	       }
	       if (base->lower_half)
		    dec_refcount(base->lower_half, max);

	       Cilk_internal_free(base, sizeof(PageBaseT));
	  } else {
	       /* free PageTableT structure */
	       for (i = 0; i < NODE_SIZE; i++) {
		    if (table->child[i])
			 dec_refcount(table->child[i], max / NODE_SIZE);
	       }
	       Cilk_internal_free(table, sizeof(PageTableT));
	  }
	  return;
     }
     Cilk_sema_signal(&table->lock);
}

/* 
 * copy page table, up to and including page z+<page>.  <table>
 * represents pages z through z+<max>-1.
 */
static volatile PageTableT *copy_local(volatile PageTableT *node,
				       int max, int page)
{
     int step, idx, off;
     int i, j;
     volatile PageBaseT *base;
     volatile PageTableT *copy;
     volatile PageBaseT *lower_half;

     CILK_ASSERT(node);
     CILK_ASSERT(page >= 0);
     CILK_ASSERT(page < max - 1);
     CILK_ASSERT(node->refcount > 0);

     if (max == LEAF_SIZE) {
	  /* make new base node */
	  base = Cilk_internal_malloc(sizeof(PageBaseT));
	  Cilk_sema_init(&base->lock);
	  base->refcount = 1;
	  for (i = 0; i <= page; i++)
	       base->pages[i] = -2;	/* 
					 * invalid - will never be 
					 * referenced 
					 */
	  for (i = page + 1; i < LEAF_SIZE; i++)
	       base->pages[i] = -1;	/* new entries for new table */
	  base->min_valid = page + 1;

	  /* make sure lower_half chain has decreasig min_valid entries */
	  lower_half = (PageBaseT *) node;
	  while (lower_half->min_valid >= page + 1)
	       lower_half = lower_half->lower_half;
	  CILK_ASSERT(lower_half->min_valid < page + 1);

	  base->lower_half = lower_half;
	  inc_refcount((volatile PageTableT *) lower_half);

	  return ((PageTableT *) base);
     }
     step = max / NODE_SIZE;	/* # of pages for each child */
     idx = page / step;		/* child # of relevant child */
     off = page % step;		/* index in relevant child   */

     copy = Cilk_internal_malloc(sizeof(PageTableT));
     Cilk_sema_init(&copy->lock);
     copy->refcount = 1;

     /* make sure children 0 through idx (inclusive) exist */
     for (i = 0; i <= idx; i++) {
	  volatile PageTableT *child = node->child[i];

	  if (child == NULL) {
	       Cilk_sema_wait(&node->lock);
	       child = node->child[i];

	       if (child == NULL) {
		    if (step == LEAF_SIZE) {
			 /* allocate new base node and set it up */
			 base = Cilk_internal_malloc(sizeof(PageBaseT));
			 Cilk_sema_init(&base->lock);
			 base->refcount = 1;
			 base->min_valid = 0;
			 base->lower_half = NULL;
			 for (j = 0; j < LEAF_SIZE; j++)
			      base->pages[j] = -1;

			 child = (PageTableT *) base;
		    } else {
			 /* allocate new table node and set it up */
			 child = Cilk_internal_malloc(sizeof(PageTableT));
			 Cilk_sema_init(&child->lock);
			 child->refcount = 1;
			 for (j = 0; j < NODE_SIZE; j++)
			      child->child[j] = NULL;
		    }
		    node->child[i] = child;
	       }
	       Cilk_sema_signal(&node->lock);
	  }
     }

     /* copy over low-indexed entries */
     for (i = 0; i < idx; i++) {
	  copy->child[i] = node->child[i];
	  inc_refcount(node->child[i]);
     }

     /* zero out high-indexed entries */
     for (i = idx + 1; i < NODE_SIZE; i++)
	  copy->child[i] = NULL;

     /* if exact, we're done */
     if (off == step - 1) {
	  copy->child[idx] = node->child[idx];
	  inc_refcount(node->child[idx]);
     } else {
	  /* recursively copy table */
	  copy->child[idx] = copy_local(node->child[idx], step, off);
     }
     return copy;
}

/* make a new page table */
volatile PageTableT *Cilk_copy_page_table(volatile PageTableT *table,
					  int page)
{
     volatile PageTableT *newtable;
     Cilk_enter_state(STATE_COPYING_PAGE_TABLE);
     CILK_ASSERT(page < MAX_PAGE - 1);

     newtable = copy_local(table, MAX_PAGE, page);
     inc_refcount(newtable);

     Cilk_exit_state(STATE_COPYING_PAGE_TABLE);
     return newtable;
}

volatile PageTableT *Cilk_dup_page_table(volatile PageTableT *table)
{
     inc_refcount(table);
     return table;
}

/* free page table */
void Cilk_free_page_table(volatile PageTableT *table)
{
     Cilk_enter_state(STATE_FREEING_PAGE_TABLE);
     CILK_ASSERT(table != NULL);
     dec_refcount(table, MAX_PAGE);
     Cilk_exit_state(STATE_FREEING_PAGE_TABLE);
}

volatile PageTableT *Cilk_create_empty_page_table(void)
{
     volatile PageTableT *table;
     int i;

     table = Cilk_internal_malloc(sizeof(PageTableT));
     Cilk_sema_init(&table->lock);
     table->refcount = 1;
     for (i = 0; i < NODE_SIZE; i++)
	  table->child[i] = NULL;

     return table;
}

/*
 * Returns first differing page between tables <t1> and <t2>.  Returns
 * <max> if no pages differ.
 */
static inline int page_tables_cmp(volatile PageTableT *t1,
				  volatile PageTableT *t2,
				  int max)
{
     int step, res;
     int i;

     CILK_ASSERT(t1 != NULL);
     CILK_ASSERT(t2 != NULL);
     if (t1 == t2)
	  return max;

     if (max == LEAF_SIZE) {
	  for (i = 0; i < LEAF_SIZE; i++) {
	       volatile PageBaseT *b1 = (volatile PageBaseT *) t1;
	       volatile PageBaseT *b2 = (volatile PageBaseT *) t2;
	       int p1, p2;

	       while (i < b1->min_valid)
		    b1 = b1->lower_half;
	       while (i < b2->min_valid)
		    b2 = b2->lower_half;

	       p1 = b1->pages[i];
	       p2 = b2->pages[i];
	       if (p1 == -1 || p2 == -1 || p1 != p2)
		    return i;
	  }
	  return max;
     }
     step = max / NODE_SIZE;	/* # of pages for each child */
     for (i = 0; i < NODE_SIZE; i++) {
	  volatile PageTableT *child1 = t1->child[i];
	  volatile PageTableT *child2 = t2->child[i];

	  if (child1 == NULL || child2 == NULL)
	       return i * step;
	  if (child1 == child2)
	       continue;

	  res = page_tables_cmp(child1, child2, step);
	  if (res != step)
	       return i * step + res;
     }
     return max;
}

int Cilk_page_tables_cmp(volatile PageTableT *t1,
			 volatile PageTableT *t2)
{
  return page_tables_cmp(t1, t2, MAX_PAGE);
}
