/*
 * $Id: timing.c,v 1.2 1996/08/27 19:28:34 randall Exp $
 */

/*
 * $Log: timing.c,v $
 * Revision 1.2  1996/08/27 19:28:34  randall
 * One pass of cleaning up.
 *
 * Revision 1.1  1996/08/26 22:17:01  athena
 * Major restructuring of the rts. Added shared and private variables.
 * Added hooks. Added cilklib.
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

/***********************************************************
 * Time-gathering stuff
 ***********************************************************/

#include <cilk-internal.h>

#if CILK_TIMING
#include <stdio.h>
extern FILE *Cilk_infofile;

/*
 * The model for time gathering is the following:
 * there is a certain set of states in which the system can
 * be. Throughout the code, ENTER_STATE(something) signals
 * that a given state is entered. EXIT_STATE(state) reverts back to
 * the previous state, and checks that the right state is being
 * exited. The time is accumulated on a per-state
 * basis. Time is cumulative, i.e. it is measured from ENTER_STATE
 * to EXIT_STATE, regardless of other states entered in between.
 * 
 * No more than STATE_DEPTH states can be nested.
 * The stack is really used only for checking purposes, but it's
 * probably a good idea to keep it.
 *
 */
#define STATE_DEPTH 256

/* also update name table below */

typedef struct {
     CACHE_LINE_ALIGNED(volatile int state_stackp);
     volatile int state_stack[STATE_DEPTH];
     volatile TimeT state_stack_begintime[STATE_DEPTH];
     volatile TimeT state_stack_begintime_wall[STATE_DEPTH];
     volatile TimeT statetime[TIME_NSTATES];
     volatile TimeT statetime_wall[TIME_NSTATES];
     volatile int stateentered[TIME_NSTATES];
} StateInfoT;

DECLARE_SHARED_VARIABLE(static StateInfoT * timer_state_info) = NULL;

/*
 * table of description of states. It is printed in
 * the same order.
 */
static const struct state_desc {
     char *name;
     int state;
} state_descriptions[] = {

     {
	  "Total running time", STATE_TOTAL
     },
     {
	  "Copying page table", STATE_COPYING_PAGE_TABLE
     },
     {
	  "Freeing page table", STATE_FREEING_PAGE_TABLE
     },
     {
	  "Stealing", STATE_STEALING
     },
     {
	  "Getting page table entry", STATE_GET_PAGE_ENTRY
     },
     {
	  "Allocating bstore page", STATE_ALLOC_BSTORE_PAGE
     },
     {
	  "Freeing bstore page", STATE_FREE_BSTORE_PAGE
     },
     {
	  "Waiting for bs lock", STATE_WAIT_BS_LOCK
     },
     {
	  "Allocating shmem", STATE_ALLOC_SHMEM
     },
     {
	  "Freeing shmem", STATE_FREE_SHMEM
     },
     {
	  "Mapping shmem pages", STATE_MAPPING_PAGE
     },
     {
	  "Doing useful work", STATE_WORKING
     },
     {
	  (char *) 0, 0
     }
};

void Cilk_enter_state(int state)
{
     int stackp;
     TimeT now = Cilk_get_process_time();
     TimeT now_wall = Cilk_get_wall_time();
     StateInfoT *info = timer_state_info + Self;
     if (timer_state_info == NULL)
	  return;
     stackp = info->state_stackp;
     CILK_ASSERT(stackp < STATE_DEPTH);
     info->state_stack[stackp] = state;
     info->state_stack_begintime[stackp] = now;
     info->state_stack_begintime_wall[stackp] = now_wall;
     info->state_stackp = stackp + 1;
     ++(info->stateentered[state]);
}

void Cilk_exit_state(int state)
{
     int stackp;
     TimeT now = Cilk_get_process_time();
     TimeT now_wall = Cilk_get_wall_time();
     StateInfoT *info = timer_state_info + Self;
     if (timer_state_info == NULL)
	  return;
     stackp = info->state_stackp - 1;
     info->state_stackp = stackp;
     CILK_ASSERT(stackp >= 0);

     CILK_ASSERT(info->state_stack[stackp] == state);
     info->statetime[state] += now - info->state_stack_begintime[stackp];
     info->statetime_wall[state] += now_wall -
	 info->state_stack_begintime_wall[stackp];
}

/* print time statistics */
void Cilk_print_time_statistics(void)
{
     int i;
     StateInfoT *info = timer_state_info;
     double worktime = 0.0;
     double totaltime = 0.0;

     fprintf(Cilk_infofile,
	     "\nSUMMARY OF TIME SPENT BY EACH PROCESSOR:\n\n"
	     " PN    Working      Total\n"
	     "-------------------------\n");
     for (i = 0; i < Cilk_active_size; i++) {
	  fprintf(Cilk_infofile, "%3d%11f%11f\n", i,
	      Cilk_process_time_to_sec(info[i].statetime[STATE_WORKING]),
	       Cilk_process_time_to_sec(info[i].statetime[STATE_TOTAL]));
	  worktime +=
	      Cilk_process_time_to_sec(info[i].statetime[STATE_WORKING]);
	  totaltime +=
	      Cilk_process_time_to_sec(info[i].statetime[STATE_TOTAL]);
     }
     fprintf(Cilk_infofile,
	     "-------------------------\n"
	     "TOT%11f%11f\n", worktime, totaltime);
}

void Cilk_print_detailed_time_statistics(void)
{
     int i;
     struct state_desc *p;
     StateInfoT *info = timer_state_info;

     fprintf(Cilk_infofile,
	     "\nDETAILS OF TIME SPENT BY EACH PROCESSOR:\n");

     for (i = 0; i < Cilk_active_size; i++) {
	  fprintf(Cilk_infofile,
		  "\n--- Time statistics for processor %d ---\n", i);

	  for (p = (struct state_desc *) state_descriptions; p->name; ++p) {
	       fprintf(Cilk_infofile,
		       "%s: %fs / wall %fs (entered %d times)\n",
		       p->name,
		   Cilk_process_time_to_sec(info[i].statetime[p->state]),
		 Cilk_wall_time_to_sec(info[i].statetime_wall[p->state]),
		       info[i].stateentered[p->state]);
	  }
     }
}

double Cilk_compute_work(void)
{
     int i;
     double work = 0.0;
     StateInfoT *info = timer_state_info;

     for (i = 0; i < Cilk_active_size; i++)
	  work += Cilk_process_time_to_sec(info[i].statetime[STATE_WORKING]);

     return work;
}

void Cilk_time_gathering_init(void)
{
     int i, j;
     StateInfoT *info;

     info = (StateInfoT *) Cilk_malloc(sizeof(StateInfoT) * Cilk_active_size);
     timer_state_info = info;

     for (i = 0; i < Cilk_active_size; i++) {
	  for (j = 0; j < TIME_NSTATES; ++j) {
	       info[i].statetime[j] = (TimeT) 0;
	       info[i].statetime_wall[j] = (TimeT) 0;
	       info[i].stateentered[j] = 0;
	  }
	  info[i].state_stackp = 0;
     }
}

#endif
